/* 
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 935 $ $LastChangedDate: 2002-09-12 20:56:28 +0200 (Thu, 12 Sep 2002) $
 *
 * Runtime environment for gont.
 *
 * Assumes compiler that allows $ in identifiers, such as GCC.
 * (It is required to have Ksi, so this ain't so important)
 */

#include <gont/runtime.h>
#include <string.h>

/* Exceptions */

#define USE_TRY_STACK 1

struct try_block {
	void *longjmpbuf[5];
	void *value;
#if !USE_TRY_STACK
	struct try_block *chain;
#endif
};


static struct try_block *top;

#if USE_TRY_STACK
#define SET_SIZE 128
struct try_block_set {
	struct try_block_set *prev;
	struct try_block blocks[SET_SIZE];
};

static struct try_block *low, *high;

static struct try_block_set *cur, *cached;

struct try_block *gont_push_try_block(void)
{
	if (top == NULL) {
		if (low == NULL) {
			cur = gont_xmalloc0(sizeof(*cur));
			low = cur->blocks;
			high = cur->blocks + SET_SIZE - 1;
		}
		
		top = low;
	} else if (++top > high) {
		if (cached) {
			cached->prev = cur;
			cur = cached;
			cached = NULL;
		} else {
			struct try_block_set *p = gont_xmalloc0(sizeof(*cur));
			p->prev = cur;
			cur = p;
		}
		
		low = cur->blocks;
		high = cur->blocks + SET_SIZE - 1;

		top = low;
	}

	return top;
}

void *gont_pop_raised_value(void)
{
	void *v = top->value;

	memset(top, '\0', sizeof(*top));

	if (--top < low) {
		cached = cur;
		cur = cur->prev;
		if (cur) {
			low = cur->blocks;
			high = cur->blocks + SET_SIZE - 1;
			top = high;
		} else {
			low = high = top = NULL;
		}
	}

	return v;
}

#else
struct try_block *gont_push_try_block(void)
{
	struct try_block *t;

	t = gont_xmalloc(sizeof(*t));

	t->value = NULL;
	t->chain = top;
	
	return (top = t);
}

void *gont_pop_raised_value(void)
{
	void *v = top->value;

	top = top->chain;

	return v;
}
#endif

void gont_pop_try_block(void)
{
	gont_pop_raised_value();
}

void gont_raise(void *v)
{
	if (top == NULL)
		gont_oops("unhandled exception");

	top->value = v;
	__builtin_longjmp(top, 1);
	/* Dummy. */
	for (;;) {}
}

extern gont_int Core$$Failure;
extern gont_int Core$$Sys_error;
extern gont_int Core$$Invalid_argument;
extern gont_int Core$$Invalid_multibyte_sequence;
extern gont_int Core$$Incomplete_multibyte_sequence;
extern gont_int Core$$Invalid_unicode_character;
extern gont_int Core$$Character_not_supported_by_encoding;

void gont_fail_with(const char *s)
{
	gont_raise(gont_make_union(Core$$Failure, gont_string_utf8(s)));
}

void gont_invalid_argument(const char *s)
{
	gont_raise(gont_make_union(Core$$Invalid_argument, gont_string_utf8(s)));
}

void gont_sys_error(gont_int err)
{
	gont_raise(gont_make_union(Core$$Sys_error, 
				   gont_string_locale(strerror(err))));
}

void gont_invalid_mbseq()
{
	gont_raise(gont_make_union(Core$$Invalid_multibyte_sequence, NULL));
}

void gont_incomplete_mbseq()
{
	gont_raise(gont_make_union(Core$$Incomplete_multibyte_sequence, NULL));
}

void gont_invalid_unicode_character()
{
	gont_raise(gont_make_union(Core$$Invalid_unicode_character, NULL));
}

void gont_character_not_supported_by_encoding()
{
	gont_raise(gont_make_union(Core$$Character_not_supported_by_encoding, 
				   NULL));
}
